Collection Mapping — SQLAlchemy 0.5.6 Documentation

# SQLAlchemy 0.5.6 Documentation

Search:
Version: 0.5.6 Last Updated: 09/13/2009 10:45:11

# Collection Mapping

This is an in-depth discussion of collection mechanics. For simple examples, see Rows that point to themselves / Mutually Dependent Rows.

Support for collections of mapped entities.

The collections package supplies the machinery used to inform the ORM of collection membership changes. An instrumentation via decoration approach is used, allowing arbitrary types (including built-ins) to be used as entity collections without requiring inheritance from a base class.

Instrumentation decoration relays membership change events to the `InstrumentedCollectionAttribute` that is currently managing the collection. The decorators observe function call arguments and return values, tracking entities entering or leaving the collection. Two decorator approaches are provided. One is a bundle of generic decorators that map function arguments and return values to events:

    from sqlalchemy.orm.collections import collection
    class MyClass(object):
        # ...

        @collection.adds(1)
        def store(self, item):
            self.data.append(item)

        @collection.removes_return()
        def pop(self):
            return self.data.pop()

The second approach is a bundle of targeted decorators that wrap appropriate append and remove notifiers around the mutation methods present in the standard Python `list`, `set` and `dict` interfaces. These could be specified in terms of generic decorator recipes, but are instead hand-tooled for increased efficiency. The targeted decorators occasionally implement adapter-like behavior, such as mapping bulk-set methods (`extend`, `update`, `__setslice__`, etc.) into the series of atomic mutation events that the ORM requires.

The targeted decorators are used internally for automatic instrumentation of entity collection classes. Every collection class goes through a transformation process roughly like so:

1. If the class is a built-in, substitute a trivial sub-class
2. Is this class already instrumented?
3. Add in generic decorators
4. Sniff out the collection interface through duck-typing
5. Add targeted decoration to any undecorated interface method

This process modifies the class at runtime, decorating methods and adding some bookkeeping properties. This isn’t possible (or desirable) for built-in classes like `list`, so trivial sub-classes are substituted to hold decoration:

    class InstrumentedList(list):
        pass

Collection classes can be specified in `relation(collection_class=)` as types or a function that returns an instance. Collection classes are inspected and instrumented during the mapper compilation phase. The collection_class callable will be executed once to produce a specimen instance, and the type of that specimen will be instrumented. Functions that return built-in types like `lists` will be adapted to produce instrumented instances.

When extending a known type like `list`, additional decorations are not generally not needed. Odds are, the extension method will delegate to a method that’s already instrumented. For example:

    class QueueIsh(list):
       def push(self, item):
           self.append(item)
       def shift(self):
           return self.pop(0)

There’s no need to decorate these methods. `append` and `pop` are already instrumented as part of the `list` interface. Decorating them would fire duplicate events, which should be avoided.

The targeted decoration tries not to rely on other methods in the underlying collection class, but some are unavoidable. Many depend on ‘read’ methods being present to properly instrument a ‘write’, for example, `__setitem__` needs `__getitem__`. “Bulk” methods like `update` and `extend` may also reimplemented in terms of atomic appends and removes, so the `extend` decoration will actually perform many `append` operations and not call the underlying method at all.

Tight control over bulk operation and the firing of events is also possible by implementing the instrumentation internally in your methods. The basic instrumentation package works under the general assumption that collection mutation will not raise unusual exceptions. If you want to closely orchestrate append and remove events with exception management, internal instrumentation may be the answer. Within your method, `collection_adapter(self)` will retrieve an object that you can use for explicit control over triggering append and remove events.

The owning object and InstrumentedCollectionAttribute are also reachable through the adapter, allowing for some very sophisticated behavior.

`sqlalchemy.orm.collections.` `attribute_mapped_collection`(attr_name)

  A dictionary-based collection type with attribute-based keying.

  Returns a MappedCollection factory with a keying based on the ‘attr_name’ attribute of entities in the collection.

  The key value must be immutable for the lifetime of the object. You can not, for example, map on foreign key values if those key values will change during the session, i.e. from None to a database-assigned integer after a session flush.

class `sqlalchemy.orm.collections.` `collection`

  Decorators for entity collection classes.

  The decorators fall into two groups: annotations and interception recipes.

  The annotating decorators (appender, remover, iterator, internally_instrumented, on_link) indicate the method’s purpose and take no arguments. They are not written with parens:

      @collection.appender
      def append(self, append): ...

  The recipe decorators all require parens, even those that take no arguments:

      @collection.adds('entity'):
      def insert(self, position, entity): ...

      @collection.removes_return()
      def popitem(self): ...

  Decorators can be specified in long-hand for Python 2.3, or with the class-level dict attribute ‘__instrumentation__’- see the source for details.

`sqlalchemy.orm.collections.` `collection_adapter`(collection)
  Fetch the CollectionAdapter for a collection.

`sqlalchemy.orm.collections.` `column_mapped_collection`(mapping_spec)

  A dictionary-based collection type with column-based keying.

  Returns a MappedCollection factory with a keying function generated from mapping_spec, which may be a Column or a sequence of Columns.

  The key value must be immutable for the lifetime of the object. You can not, for example, map on foreign key values if those key values will change during the session, i.e. from None to a database-assigned integer after a session flush.

`sqlalchemy.orm.collections.` `mapped_collection`(keyfunc)

  A dictionary-based collection type with arbitrary keying.

  Returns a MappedCollection factory with a keying function generated from keyfunc, a callable that takes an entity and returns a key value.

  The key value must be immutable for the lifetime of the object. You can not, for example, map on foreign key values if those key values will change during the session, i.e. from None to a database-assigned integer after a session flush.